/*
 * Copyright (c) 2000 Greg Haerr <greg@censoft.com>
 * Copyright (c) 1991 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * DYNAMICREGIONS WndSetClipWindow
 */
#include "windows.h"
#include "wintern.h"

/*
 * Set the clip rectangles for a window taking into account other
 * windows that may be obscuring it.  The windows that may be obscuring
 * this one are the siblings of each direct ancestor which are higher
 * in priority than those ancestors.  Also, each parent limits the visible
 * area of the window.
 */
void
WndSetClipWindow(HDC hdc)
{
	HWND		wp = hdc->hwnd;
	HWND		pwp;		/* parent window */
	HWND		sibwp;		/* sibling windows */
	COORD		diff;		/* difference in coordinates */
	PRECT		prc;		/* client or window rectangle*/
	MWCLIPREGION	*vis, *r;
	COORD		x, y, width, height;

	if (wp->unmapcount)
		return;

	/*
	 * Start with the rectangle for the complete window.
	 * We will then cut pieces out of it as needed.
	 */
	prc = WndIsClientDC(hdc)? &wp->clirect: &wp->winrect;
	x = prc->left;
	y = prc->top;
	width = prc->right - prc->left;
	height = prc->bottom - prc->top;

#if UPDATEREGIONS
	if(wp->nRcUpdate == 0)
		goto cliptoparent;

	/*
	 * If update regions are used, replace window rectangle
	 * with series of update rectangles.  These rectangles
	 * will then be cropped back for client DC's.  Each
	 * update rectangle will then be used as a "view" of the
	 * complete window for later sibling clipping.
	 */
	for(count=0; count<wp->nRcUpdate; ++count) {
		prc = &wp->RcUpdate[count];
		clip->x = prc->left + wp->winrect.left;
		clip->y = prc->top + wp->winrect.top;
		clip->width = prc->right - prc->left;
		clip->height = prc->bottom - prc->top;

		if(WndIsClientDC(hdc)) {
			clip->x = max(clip->x, wp->clirect.left);
			clip->y = max(clip->y, wp->clirect.top);
			clip->width = min(clip->width,
				wp->clirect.right-wp->clirect.left);
			clip->height = min(clip->height,
				wp->clirect.bottom-wp->clirect.top);
		}

/*printf("win %d,%d,%d,%d clip %d,%d,%d,%d count %d, %d\n",
wp->winrect.left, wp->winrect.top, wp->winrect.right-wp->winrect.left,
wp->winrect.bottom-wp->winrect.top, clip->x, clip->y, clip->width,
clip->height, count, wp->nRcUpdate);*/
#endif

	/*
	 * First walk upwards through all parent windows,
	 * and restrict the visible part of this window to the part
	 * that shows through all of those parent windows client areas.
	 */
#if UPDATEREGIONS
cliptoparent:
#endif
	pwp = wp;
	while (pwp != rootwp) {
		pwp = pwp->parent;

		diff = pwp->clirect.left - x;
		if (diff > 0) {
			width -= diff;
			x = pwp->clirect.left;
		}

		diff = pwp->clirect.right - (x + width);
		if (diff < 0)
			width += diff;

		diff = pwp->clirect.top - y;
		if (diff > 0) {
			height -= diff;
			y = pwp->clirect.top;
		}

		diff = pwp->clirect.bottom - (y + height);
		if (diff < 0)
			height += diff;
	}

	/*
	 * If the window is completely clipped out of view, then
	 * set the clipping region to indicate that.
	 */
	if (width <= 0 || height <= 0) {
#if !UPDATEREGIONS
		GdSetClipRegion(hdc->psd, NULL);
		return;
#endif
	} 
#if UPDATEREGIONS
		else
			++clip;
	} /* for loop*/
	if(clip == cliprects) {
		GdSetClipRegion(hdc->psd, NULL);
		return;
	}
#endif

	vis = GdAllocRectRegion(x, y, x+width, y+height);
	/* FIXME: intersect with screen here?*/

	/*
	 * Now examine all windows that obscure this window, and
	 * for each obscuration, break up the clip rectangles into
	 * the smaller pieces that are still visible.  The windows
	 * that can obscure us are the earlier siblings of all of
	 * our parents. When clipping the root window, search all children.
 	 */
	pwp = wp;
	while (pwp != NULL) {
		wp = pwp;
		pwp = wp->parent;
		if(!pwp) {
			/* We're clipping the root window*/
			if(hdc->flags & DCX_CLIPCHILDREN)
				/* start with root's children*/
				sibwp = rootwp->children;
			else sibwp = NULL;	/* no search*/
			wp = NULL;		/* search all root's children*/
		} else {
			if(hdc->flags & DCX_CLIPSIBLINGS)
				sibwp = pwp->children;
			else sibwp = wp;	/* no search*/
		}
		for (; sibwp != wp; sibwp = sibwp->siblings) {
			if (sibwp->unmapcount)
				continue;

			r = GdAllocRectRegionIndirect(&sibwp->winrect);
			GdSubtractRegion(vis, vis, r);
			GdDestroyRegion(r);
		}

		/* if not clipping the root window, stop when you reach it*/
		if(pwp == rootwp)
			break;
	}

	/*
	 * If not the root window and we're going to be drawing
	 * in the client area, clip all children.  This is
	 * required for non-special paint handling for child windows.
	 * Non-client dc's don't clip children in order to get
	 * proper border clipping in the case of border-clipped children.
	 */
	wp = hdc->hwnd;
	if(wp != rootwp && WndIsClientDC(hdc)) {
		for (sibwp=wp->children; sibwp; sibwp = sibwp->siblings) {
			if (sibwp->unmapcount)
				continue;

			r = GdAllocRectRegionIndirect(&sibwp->winrect);
			GdSubtractRegion(vis, vis, r);
			GdDestroyRegion(r);
		}
	}

	/*
	 * Intersect with user region, if set.
	 */
	if (hdc->region)
		GdIntersectRegion(vis, vis, hdc->region->rgn);
	/*
	 * Set the clip region.
	 */
	GdSetClipRegion(hdc->psd, vis);
}
